package it.grid.storm.space;

import java.io.File;
import java.util.Date;
import java.util.concurrent.Callable;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

public class CallableDU implements Callable<DUResult> {

    private static final Logger LOG = LoggerFactory.getLogger(CallableDU.class);
    private String rootPath;
    private long timeout;
    private Date creationTime;
    private Date startTime;
    private Date endTime;
    private ExitCode cmdResult;
    private static final String POISON_PILL = "poison-pill";

    public enum ExitCode {
        SUCCESS(0), IO_ERROR(1), INTERRUPTED(2), TIMEOUT(3), UNDEFINED(-1), POISON_PILL(99);

        private int code;

        ExitCode(int c) {
            code = c;
        }

        public int getCode() {
            return code;
        }

        static ExitCode getExitCode(int code) {
            ExitCode result = ExitCode.UNDEFINED;
            for (ExitCode ec : ExitCode.values()) {
                if (ec.getCode() == code) {
                    result = ec;
                    break;
                }
            }
            return result;
        }
    }

    public CallableDU(String rootPath, long timeOutInSeconds) {
        super();
        this.rootPath = rootPath;
        this.timeout = timeOutInSeconds * 1000;
        this.creationTime = new Date(System.currentTimeMillis());
    }

    public static CallableDU getPoisonPill() {
        return new CallableDU(POISON_PILL, 1000);
    }
    
    private boolean isWindows() {
        String os = System.getProperty("os.name").toLowerCase();
        // windows
        return (os.indexOf("win") >= 0);
    }

    private String getDULocalCommand() {
        String result = "du"; // default is Linux
        if (isWindows()) {
            result = "C:\\cygwin\\bin\\du.exe";
            File cf = new File(result);
            if (!(cf.exists())) {
                LOG.error("Unable to find DU command.");
                result = null;
            }
        }
        return result;
    }

    /**
     * @return the rootPath
     */
    public final String getRootPath() {
        return rootPath;
    }

    /**
     * @return the creationTime
     */
    public final Date getCreationTime() {
        return creationTime;
    }

    /**
     * @return the startTime
     */
    public final Date getStartTime() {
        return startTime;
    }

    /**
     * @return the endTime
     */
    public final Date getEndTime() {
        return endTime;
    }

    /**
     * @return the cmdResult
     */
    public final ExitCode getCmdResult() {
        return cmdResult;
    }

    public DUResult call() throws Exception {
        long size = 0;
        long startT = System.currentTimeMillis();
        this.startTime = new Date(startT);
        String command = getDULocalCommand();
        DUResult result;

        if (this.getRootPath().equals(POISON_PILL)) {
            result = new DUResult(0, POISON_PILL , startTime, 0, ExitCode.POISON_PILL);
            return result;
        }
        /* Added -b option to obtain the size in bytes */
        ExecCommand ec = new ExecCommand(command + " -s -b " + this.rootPath, this.timeout);
        cmdResult = ExitCode.getExitCode(ec.runCommand());
        LOG.debug("Command result: " + cmdResult);
        String output = ec.getOutput();
        LOG.debug(" Output: '" + output + "'");
        if (output != null) {
            String[] outputArray = output.split("\\s");
            for (int i = 0; i < outputArray.length; i++) {
                LOG.trace("outputArray[" + i + "]=" + outputArray[i]);
            }
            try {
                size = Long.valueOf(outputArray[0]).longValue();
            } catch (NumberFormatException nfe) {
                LOG.error("Unable to retrieve the disk usage of '" + this.rootPath + "'. " + nfe.getMessage());
            }
        }
        long endT = System.currentTimeMillis();
        long durationT = endT - startT;
        this.endTime = new Date(endT);

        result = new DUResult(size, rootPath, startTime, durationT, cmdResult);

        return result;
    }

}
