package it.grid.storm.space;

import it.grid.storm.concurrency.NamedThread;

import java.io.BufferedReader;
import java.io.IOException;
import java.io.InputStreamReader;
import java.io.PrintWriter;
import java.util.concurrent.Semaphore;
import java.util.concurrent.TimeoutException;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

public class ExecCommand {

    private static final Logger LOG = LoggerFactory.getLogger(ExecCommand.class);

    private Semaphore outputSem;
    private String output;
    private Semaphore errorSem;
    private String error;
    private Process process;
    private long timeout = 0;
    private String command = null;
    private String input = null;

    private class InputWriter extends Thread {
        private String input;
        private PrintWriter pw;

        public InputWriter(String input) {
            this.input = input;
        }

        public void run() {
            try {
                pw = new PrintWriter(process.getOutputStream());
                pw.println(input);
                pw.flush();
                if (pw.checkError()) {
                    LOG.warn("Something wrong during passing Command to the Process");
                    pw.close();
                }
            } catch (Throwable t) {
                if (pw != null) {
                    pw.flush();
                }
            }
        }
    }

    private class OutputReader extends Thread {
        public OutputReader() {
            try {
                outputSem = new Semaphore(1);
                outputSem.acquire();
            } catch (InterruptedException e) {
                LOG.warn("Interrupt occours when retrieve output form the execution on a native command. "
                        + e.getMessage());
            }
        }

        public void run() {
            BufferedReader isr;
            try {
                StringBuffer readBuffer = new StringBuffer();
                isr = new BufferedReader(new InputStreamReader(process.getInputStream()));
                String buff;
                while ((buff = isr.readLine()) != null) {
                    readBuffer.append(buff);
                    LOG.debug(" - output-reader: " + buff);
                }
                output = readBuffer.toString();
                outputSem.release();
                if (isr != null) {
                    isr.close();
                }
            } catch (IOException e) {
                LOG.warn("IO Exception occours when retrieve output form the execution on a native command. "
                        + e.getMessage());
            }
        }
    }

    private class ErrorReader implements Runnable {
        public ErrorReader() {
            try {
                errorSem = new Semaphore(1);
                errorSem.acquire();
            } catch (InterruptedException e) {
                LOG.warn("Interrupt occours when retrieve ERROR output form the execution on a native command. "
                        + e.getMessage());
            }
        }

        public void run() {
            BufferedReader isr = null;
            try {
                StringBuffer readBuffer = new StringBuffer();
                isr = new BufferedReader(new InputStreamReader(process.getErrorStream()));
                String buff;
                while ((buff = isr.readLine()) != null) {
                    readBuffer.append(buff);
                    LOG.debug(" - error-reader: " + buff);
                }
                error = readBuffer.toString();
                errorSem.release();
                if (isr != null) {
                    isr.close();
                }
            } catch (IOException e) {
                LOG.warn("IO Exception occours when retrieve ERROR output form the execution on a native command. "
                        + e.getMessage());
            }
            if (error.length() > 0)
                LOG.warn("Error returned by native command: " + error);
        }
    }

    /**
     * Constructor
     * 
     * @param command
     * @param input
     * @param timeout
     */
    public ExecCommand(final String command, final String input, final long timeout) {
        this.command = command;
        this.input = input;
        this.timeout = timeout;
    }

    /**
     * Constructor
     * 
     * @param command
     * @param input
     */
    public ExecCommand(final String command, final String input) {
        this.command = command;
        this.input = input;
        this.timeout = -1;
    }

    /**
     * Constructor
     * 
     * @param command
     * @param timeout
     */
    public ExecCommand(final String command, final long timeout) {
        this.command = command;
        this.input = "";
        this.timeout = timeout;
    }

    /**
     * Constructor
     * 
     * @param command
     */
    public ExecCommand(final String command) {
        this.command = command;
        this.input = "";
        this.timeout = -1;
    }

    /**
	 * 
	 */
    public int runCommand() {
        int result = 0; // zero means success
        try {
            Runtime runtime = Runtime.getRuntime();
            process = runtime.exec(command);
            LOG.debug("Running command: '" + command + "'");
            // Setting up I/O
            if ((input != null) && (!input.isEmpty())) {
                new NamedThread(new InputWriter(input), "ExecCommand-InputWriter").start();
            }
            new NamedThread(new OutputReader(), "ExecCommand-OutputReader").start();
            new NamedThread(new ErrorReader(), "ExecCommand-ErrorReader").start();
            
            // Setting up the worker
            Worker worker = new Worker(process);
            result = exec(worker, timeout);
            LOG.debug("Command result with value: " + result);
            // Wait for the completion of task
            process.waitFor();
        } catch (IOException e) {
            LOG.warn("IO Exception occours during the execution of a native command " + command + ". "
                    + e.getMessage());
            result = 1;
        } catch (InterruptedException e) {
            LOG.warn("Interrupted Exception occours during the execution of a native command " + command
                    + ". " + e.getMessage());
            result = 2;
        } catch (TimeoutException e) {
            LOG.warn("Native command " + command + " was in TimeOut.");
            result = 3;
        } catch (NullPointerException npe) {
            LOG.error("The command to execute is NULL! " + npe);
            result = 4;
        } catch (SecurityException se) {
            LOG.error("The program have Security limitation to execute the command '" + command + "'." + se);
            result = 5;
        } catch (IllegalArgumentException iae) {
            LOG.error("The command to execute is EMPTY or ILLEGAL ('" + command + "')! " + iae);
            result = 6;
        }
        return result;
    }

    private int exec(Worker worker, long timeout) throws TimeoutException, InterruptedException {
        worker.start();
        try {
            worker.join(timeout);
            if (worker.exit != null)
                return worker.exit;
            else
                throw new TimeoutException();
        } catch (InterruptedException ex) {
            worker.interrupt();
            Thread.currentThread().interrupt();
            throw ex;
        } finally {
            process.destroy();
        }
    }

    private static class Worker extends Thread {
        private final Process process;
        private Integer exit;

        private Worker(Process process) {
            this.process = process;
        }

        public void run() {
            try {
                exit = process.waitFor();
            } catch (InterruptedException ignore) {
                return;
            }
        }
    }

    public String getOutput() {
        try {
            outputSem.acquire();
        } catch (InterruptedException e) {
            LOG.warn("Interrupt occours when acquiring Semaphore the execution on a native command. "
                    + e.getMessage());
        }
        String value = output;
        outputSem.release();
        return value;
    }

    public String getError() {
        try {
            errorSem.acquire();
        } catch (InterruptedException e) {
            LOG.warn("Interrupt occours when acquiring Semaphore the execution on a native command. "
                    + e.getMessage());
        }
        String value = error;
        errorSem.release();
        return value;
    }

}
